#include <tos.h>
#include <aes.h>
#include <stdarg.h>
#include <stdio.h>
#include <mint/sysvars.h>
#define RSC_NAMED_FUNCTIONS 1
#define hfix_objs(a, b, c)
#include "tos_fuji.rsh"

#ifndef FALSE
# define FALSE 0
# define TRUE  1
#endif

typedef enum {
	COUNTRY_US =  0,	/* USA */
	COUNTRY_DE =  1,	/* Germany */
	COUNTRY_FR =  2,	/* France */
	COUNTRY_UK =  3,	/* United Kingdom */
	COUNTRY_ES =  4,	/* Spain */
	COUNTRY_IT =  5,	/* Italy */
	COUNTRY_SE =  6,	/* Sweden */
	COUNTRY_SF =  7,	/* Switzerland (French) */
	COUNTRY_SG =  8,	/* Switzerland (German) */
	COUNTRY_TR =  9,	/* Turkey */
	COUNTRY_FI = 10,	/* Finland */
	COUNTRY_NO = 11,	/* Norway */
	COUNTRY_DK = 12,	/* Denmark */
	COUNTRY_SA = 13,	/* Saudi Arabia */
	COUNTRY_NL = 14,	/* Holland */
	COUNTRY_CZ = 15,	/* Czech Republic */
	COUNTRY_HU = 16,	/* Hungary */
	COUNTRY_PL = 17,	/* Poland */
	COUNTRY_LT = 18,	/* Lithuania */
	COUNTRY_RU = 19,	/* Russia */
	COUNTRY_EE = 20,	/* Estonia */
	COUNTRY_BY = 21,	/* Belarus */
	COUNTRY_UA = 22,	/* Ukraine */
	COUNTRY_SK = 23,	/* Slovak Republic */
	COUNTRY_RO = 24,	/* Romania */
	COUNTRY_BG = 25,	/* Bulgaria */
	COUNTRY_SI = 26,	/* Slovenia */
	COUNTRY_HR = 27,	/* Croatia */
	COUNTRY_RS = 28,	/* Serbia */
	COUNTRY_ME = 29,	/* Montenegro */
	COUNTRY_MK = 30,	/* Macedonia */
	COUNTRY_GR = 31,	/* Greece */
	COUNTRY_LV = 32,	/* Latvia */
	COUNTRY_IL = 33,	/* Israel */
	COUNTRY_ZA = 34,	/* South Africa */
	COUNTRY_PT = 35,	/* Portugal */
	COUNTRY_BE = 36,	/* Belgium */
	COUNTRY_JP = 37,	/* Japan */
	COUNTRY_CN = 38,	/* China */
	COUNTRY_KR = 39,	/* Korea */
	COUNTRY_VN = 40,	/* Vietnam */
	COUNTRY_IN = 41,	/* India */
	COUNTRY_IR = 42,	/* Iran */
	COUNTRY_MN = 43,	/* Mongolia */
	COUNTRY_NP = 44,	/* Nepal */
	COUNTRY_LA = 45,	/* Lao People's Democratic Republic */
	COUNTRY_KH = 46,	/* Cambodia */
	COUNTRY_ID = 47,	/* Indonesia */
	COUNTRY_BD = 48,	/* Bangladesh */

	COUNTRY_ANY = 127
} countrycode;

#define ONE_KB 0x400UL
#define ONE_MB 0x100000UL

static _WORD gl_wchar, gl_hchar;
static _WORD aes_handle;
static _WORD dummy;

typedef struct version {
	short tos_version;
	countrycode language;
	unsigned long filesize;
	unsigned long logoaddr;
	unsigned long testaddr;
	unsigned short testwords[6];
};

static struct version const versions[] = {
	{ 0x206, COUNTRY_US,  0x40000UL, 0x03257EUL, 0x00075cUL, { 0x08b8, 0x0005, 0xfa07, 0x207c, 0x00e3, 0x257E } },
	{ 0x206, COUNTRY_DE,  0x40000UL, 0x0325AEUL, 0x00075cUL, { 0x08b8, 0x0005, 0xfa07, 0x207c, 0x00e3, 0x25AE } },
	{ 0x206, COUNTRY_FR,  0x40000UL, 0x0325CCUL, 0x00075cUL, { 0x08b8, 0x0005, 0xfa07, 0x207c, 0x00e3, 0x25CC } },
	{ 0x206, COUNTRY_UK,  0x40000UL, 0x03257EUL, 0x00075cUL, { 0x08b8, 0x0005, 0xfa07, 0x207c, 0x00e3, 0x257E } },
	{ 0x206, COUNTRY_ES,  0x40000UL, 0x0325DAUL, 0x00075cUL, { 0x08b8, 0x0005, 0xfa07, 0x207c, 0x00e3, 0x25DA } },
	{ 0x206, COUNTRY_IT,  0x40000UL, 0x0325BCUL, 0x00075cUL, { 0x08b8, 0x0005, 0xfa07, 0x207c, 0x00e3, 0x25BC } },
	{ 0x206, COUNTRY_SE,  0x40000UL, 0x0325CCUL, 0x00075cUL, { 0x08b8, 0x0005, 0xfa07, 0x207c, 0x00e3, 0x25CC } },
	{ 0x206, COUNTRY_SF,  0x40000UL, 0x0325CCUL, 0x00075cUL, { 0x08b8, 0x0005, 0xfa07, 0x207c, 0x00e3, 0x25CC } },
	{ 0x206, COUNTRY_SG,  0x40000UL, 0x0325AEUL, 0x00075cUL, { 0x08b8, 0x0005, 0xfa07, 0x207c, 0x00e3, 0x25AE } },
	{ 0x206, COUNTRY_CZ,  0x40000UL, 0x0325AEUL, 0x00075cUL, { 0x08b8, 0x0005, 0xfa07, 0x207c, 0x00e3, 0x25AE } },
	{ 0x206, COUNTRY_RU,  0x40000UL, 0x0325AEUL, 0x00075cUL, { 0x08b8, 0x0005, 0xfa07, 0x207c, 0x00e3, 0x25AE } },

	{ 0x208, COUNTRY_US,  0x40000UL, 0x03252EUL, 0x0006a4UL, { 0x08b8, 0x0005, 0xfa07, 0x207c, 0x00e3, 0x252E } },
	{ 0x208, COUNTRY_DE,  0x40000UL, 0x03255EUL, 0x0006a4UL, { 0x08b8, 0x0005, 0xfa07, 0x207c, 0x00e3, 0x255E } },
	{ 0x208, COUNTRY_FR,  0x40000UL, 0x03257CUL, 0x0006a4UL, { 0x08b8, 0x0005, 0xfa07, 0x207c, 0x00e3, 0x257C } },
	{ 0x208, COUNTRY_UK,  0x40000UL, 0x03252EUL, 0x0006a4UL, { 0x08b8, 0x0005, 0xfa07, 0x207c, 0x00e3, 0x252E } },

	{ 0x306, COUNTRY_US,  0x80000UL, 0x035fe8UL, 0x000744UL, { 0x3111, 0x4e7b, 0x0002, 0x207c, 0x00e3, 0x5fe8 } },
	{ 0x306, COUNTRY_DE,  0x80000UL, 0x036018UL, 0x000744UL, { 0x3111, 0x4e7b, 0x0002, 0x207c, 0x00e3, 0x6018 } },
	{ 0x306, COUNTRY_FR,  0x80000UL, 0x036036UL, 0x000744UL, { 0x3111, 0x4e7b, 0x0002, 0x207c, 0x00e3, 0x6036 } },
	{ 0x306, COUNTRY_UK,  0x80000UL, 0x035fe8UL, 0x000744UL, { 0x3111, 0x4e7b, 0x0002, 0x207c, 0x00e3, 0x5fe8 } },
	{ 0x306, COUNTRY_ES,  0x80000UL, 0x036044UL, 0x000744UL, { 0x3111, 0x4e7b, 0x0002, 0x207c, 0x00e3, 0x6044 } },
	{ 0x306, COUNTRY_IT,  0x80000UL, 0x03602cUL, 0x000744UL, { 0x3111, 0x4e7b, 0x0002, 0x207c, 0x00e3, 0x602c } },
	{ 0x306, COUNTRY_SE,  0x80000UL, 0x036036UL, 0x000744UL, { 0x3111, 0x4e7b, 0x0002, 0x207c, 0x00e3, 0x6036 } },
	{ 0x306, COUNTRY_SF,  0x80000UL, 0x036036UL, 0x000744UL, { 0x3111, 0x4e7b, 0x0002, 0x207c, 0x00e3, 0x6036 } },
	{ 0x306, COUNTRY_SG,  0x80000UL, 0x036018UL, 0x000744UL, { 0x3111, 0x4e7b, 0x0002, 0x207c, 0x00e3, 0x6018 } },
	{ 0x306, COUNTRY_CZ,  0x80000UL, 0x036018UL, 0x000744UL, { 0x3111, 0x4e7b, 0x0002, 0x207c, 0x00e3, 0x6018 } },
	{ 0x306, COUNTRY_RU,  0x80000UL, 0x036018UL, 0x000744UL, { 0x3111, 0x4e7b, 0x0002, 0x207c, 0x00e3, 0x6018 } },

	{ 0x404, COUNTRY_ANY, 0x80000UL, 0x049434UL, 0x049424UL, { 0x8765, 0x4321, 0x0000, 0xf99c, 0x00e1, 0xf38a } },
};
#define NUM_VERSIONS ((int)(sizeof(versions) / sizeof(versions[0])))

#ifndef PATH_MAX
#define PATH_MAX 1024
#endif
static char bmppath[PATH_MAX];
static char abspath[PATH_MAX];
static char tospath[PATH_MAX];
static unsigned char *tos_img;
static unsigned long tos_file_size;
static unsigned short tos_version;
static int is_emutos;

typedef struct {
	/* file header */
	unsigned char magic[2];		/* BM */
	unsigned char filesize[4];
	unsigned char xHotSpot[2];     
	unsigned char yHotSpot[2];
	unsigned char offbits[4];		/* offset to data */

	/* info header */
	union {
		struct {
			unsigned char hsize[4];			/* size of info header (40) */
			unsigned char width[4];			/* width in pixels */
			unsigned char height[4];		/* height in pixels */
			unsigned char planes[2];		/* always 1 */
			unsigned char bitcount[2];		/* bits per pixel: 1, 4, 8, 24 or 32 */
			unsigned char compression[4];	/* compression method */
#define BMP_RGB 0
#define BMP_RLE8 1
#define BMP_RLE4 2
#define BMP_BITFIELDS 3
#define BMP_JPEG 4
#define BMP_PNG 5
			unsigned char sizeImage[4];		/* size of data */
			unsigned char xPelsPerMeter[4];
			unsigned char yPelsPerMeter[4];
			unsigned char clrUsed[4];		/* # of colors used */
			unsigned char clrImportant[4];	/* # of important colors */
			unsigned char RedMask[4];
			unsigned char GreenMask[4];
			unsigned char BlueMask[4];
			unsigned char AlphaMask[4];
		} bitmapinfoheader;
		struct {
			unsigned char hsize[4];			/* size of info header (12) */
			unsigned char width[2];			/* width in pixels */
			unsigned char height[2];		/* height in pixels */
			unsigned char planes[2];		/* always 1 */
			unsigned char bitcount[2];		/* bits per pixel: 1, 4, 8, 24 or 32 */
		} bitmapcoreheader;
	} bmp_info_header;
} BMP_HEADER;

static BMP_HEADER bmpheader;

#define ICON_WIDTH 96
#define ICON_HEIGHT 86
#define ICON_SIZE ((ICON_WIDTH / 8) * ICON_HEIGHT)

static unsigned char icondata[(ICON_WIDTH / 8) * ICON_HEIGHT];
static int icon_loaded;


/* -------------------------------------------------------------------------- */

static _WORD do_dialog(_WORD num, _WORD screenx, _WORD screeny)
{
	OBJECT *tree = rs_trindex[num];
	GRECT gr;
	_WORD ret;

	wind_update(BEG_UPDATE);
	form_center_grect(tree, &gr);
	if (screenx != 0 && screeny != 0)
	{
		_WORD dx = tree[ROOT].ob_x - gr.g_x;
		_WORD dy = tree[ROOT].ob_y - gr.g_y;
		tree[ROOT].ob_x = screenx;
		tree[ROOT].ob_y = screeny;
		gr.g_x = screenx - dx;
		gr.g_y = screeny - dy;
	}
	form_dial_grect(FMD_START, &gr, &gr);
	objc_draw_grect(tree, ROOT, MAX_DEPTH, &gr);
	ret = form_do(tree, ROOT);
	ret &= 0x7fff;
	tree[ret].ob_state &= ~OS_SELECTED;
	form_dial_grect(FMD_FINISH, &gr, &gr);
	wind_update(END_UPDATE);
	evnt_timer(0);
	return ret;
}


static long get_rom_addr(void)
{
	OSHEADER *sysbase = *((OSHEADER **)0x4f2);
	sysbase = sysbase->os_base;
	return (long)sysbase;
}


static _WORD do_alert(_WORD id, ...)
{
	va_list args;
	char buf[256];
	
	va_start(args, id);
	vsprintf(buf, rs_frstr[id], args);
	va_end(args);
	return form_alert(1, buf);
}


static int check_tosversion(const OSHEADER *hdr, unsigned long filesize)
{
	int i, j;
	const unsigned short *test;
	unsigned short version;
	
	for (i = 0; i < NUM_VERSIONS; i++)
	{
		version = hdr->os_version;
		if (version == 0x206 && hdr->os_dosdate == 0x186a)
			version = 0x208; /* ST-Book TOS */
		if (version != versions[i].tos_version)
			continue;
		if (((hdr->os_conf >> 1) & 0x7f) != versions[i].language)
			continue;
		if (filesize != versions[i].filesize)
		{
			do_alert(AL_WRONG_ROMSIZE);
			return -1;
		}
		is_emutos = ((long)hdr->p_rsv2) == 0x45544f53L;
		tos_version = version;
		test = (const unsigned short *)((const char *)hdr + versions[i].testaddr);
		for (j = 0; j < 6; j++)
		{
			if (test[j] != versions[i].testwords[j])
			{
				if (do_alert(AL_TEST_FAILED) != 1)
					return -1;
			}
		}
		return i;
	}
	do_alert(AL_UNKNOWN_TOS);
	return -1;
}


static int select_file(char *path, const char *name, const char *title)
{
	size_t len;
	char *p, *p2;
	_WORD button;
	char filename[PATH_MAX];
	
	if (path[0] == '\0')
	{
		int drv = Dgetdrv();
		path[0] = drv + 'A';
		path[1] = ':';
		Dgetpath(path + 2, drv + 1);
		len = strlen(path);
		if (len > 0 && path[len - 1] != '\\' && path[len - 1] != '/')
			strcat(path, "\\");
	} else
	{
		p = strrchr(path, '\\');
		p2 = strrchr(path, '/');
		if (p == NULL || p2 > p)
			p = p2;
		if (p)
			p[1] = '\0';
	}
	*filename = '\0';
	strcat(path, name);
	button = 0;
	if (!fsel_exinput(path, filename, &button, title) || button == 0 || filename[0] == '\0')
		return FALSE;
	p = strrchr(path, '\\');
	p2 = strrchr(path, '/');
	if (p == NULL || p2 > p)
		p = p2;
	if (p)
		p[1] = '\0';
	strcat(path, filename);
	return TRUE;
}


static void put_long(unsigned char *dst, unsigned long val)
{
	*dst++ = (unsigned char)val;
	val >>= 8;
	*dst++ = (unsigned char)val;
	val >>= 8;
	*dst++ = (unsigned char)val;
	val >>= 8;
	*dst = (unsigned char)val;
}


static void put_word(unsigned char *dst, unsigned short val)
{
	*dst++ = (unsigned char)val;
	val >>= 8;
	*dst = (unsigned char)val;
}


static unsigned long get_long(const unsigned char *src)
{
	unsigned long val;
	val = src[3];
	val <<= 8;
	val |= src[2];
	val <<= 8;
	val |= src[1];
	val <<= 8;
	val |= src[0];
	return val;
}


static unsigned short get_word(const unsigned char *src)
{
	unsigned short val;
	val = src[1];
	val <<= 8;
	val |= src[0];
	return val;
}


static int write_bmpfile(const unsigned char *data)
{
	unsigned long headlen;
	unsigned long len;
	int fd;
	unsigned char palette[2][4];
	unsigned long rowsize;
	int y;

	fd = (int)Fcreate(bmppath, 0);
	if (fd < 0)
	{
		do_alert(AL_FILE_CREATE);
		return FALSE;
	}
	memset(&bmpheader, 0, sizeof(bmpheader));
	headlen = 14 + 40 + 2 * 4;
	len = headlen + ICON_SIZE;
	
	bmpheader.magic[0] = 'B';
	bmpheader.magic[1] = 'M';
	put_long(bmpheader.filesize, len);
	put_long(bmpheader.offbits, headlen);
	put_long(bmpheader.bmp_info_header.bitmapinfoheader.hsize, 40);
	put_word(bmpheader.bmp_info_header.bitmapinfoheader.width, ICON_WIDTH);
	put_word(bmpheader.bmp_info_header.bitmapinfoheader.height, ICON_HEIGHT);
	put_word(bmpheader.bmp_info_header.bitmapinfoheader.planes, 1);
	put_word(bmpheader.bmp_info_header.bitmapinfoheader.bitcount, 1);
	put_long(bmpheader.bmp_info_header.bitmapinfoheader.clrUsed, 2);
	put_long(palette[0], 0xffffffL);
	put_long(palette[1], 0x000000L);
	Fwrite(fd, 14 + 40, &bmpheader);
	Fwrite(fd, sizeof(palette), palette);
	rowsize = ICON_WIDTH / 8;
	data += ICON_HEIGHT * rowsize;
	for (y = 0; y < ICON_HEIGHT; y++)
	{
		data -= rowsize;
		Fwrite(fd, rowsize, data);
	}

	Fclose(fd);
	
	return TRUE;
}


static int read_bmpfile(void)
{
	unsigned long headlen;
	int fd;
	unsigned char palette[2][4];
	unsigned long rowsize;
	int y;
	long width;
	long height;
	int topdown;
	unsigned short planes, bitcount;
	int mapentrysize;
	long clrused;
	int i;
	unsigned char *data;
	
	fd = (int)Fopen(bmppath, FO_READ);
	if (fd < 0)
	{
		do_alert(AL_FILE_OPEN);
		return FALSE;
	}
	memset(&bmpheader, 0, sizeof(bmpheader));
	headlen = 14 + 40;
	Fread(fd, headlen, &bmpheader);
	headlen = get_long(bmpheader.bmp_info_header.bitmapinfoheader.hsize);
	if (bmpheader.magic[0] != 'B' ||
		bmpheader.magic[1] != 'M' ||
		(headlen != 12 && headlen != 40 && headlen != 56 && headlen != 64 && headlen != 108 && headlen != 124) ||
		(headlen > 12 && get_long(bmpheader.bmp_info_header.bitmapinfoheader.compression) != 0))
	{
		Fclose(fd);
		do_alert(AL_UNKNOWN_FORMAT);
		return FALSE;
	}
	if (headlen == 12)
	{
		width = get_word(bmpheader.bmp_info_header.bitmapcoreheader.width);
		height = (short)get_word(bmpheader.bmp_info_header.bitmapcoreheader.height);
		planes = get_word(bmpheader.bmp_info_header.bitmapcoreheader.planes);
		bitcount = get_word(bmpheader.bmp_info_header.bitmapcoreheader.bitcount);
		mapentrysize = 3;
		clrused = 0;
	} else
	{
		width = get_long(bmpheader.bmp_info_header.bitmapinfoheader.width);
		height = get_long(bmpheader.bmp_info_header.bitmapinfoheader.height);
		planes = get_word(bmpheader.bmp_info_header.bitmapinfoheader.planes);
		bitcount = get_word(bmpheader.bmp_info_header.bitmapinfoheader.bitcount);
		mapentrysize = 4;
		clrused = get_long(bmpheader.bmp_info_header.bitmapinfoheader.clrUsed);
	}
	topdown = FALSE;
	if (height < 0)
	{
		topdown = TRUE;
		height = -height;
	}
	if (width != ICON_WIDTH || height != ICON_HEIGHT ||
		planes != 1 || bitcount != 1)
	{
		Fclose(fd);
		do_alert(AL_UNSUPPORTED_FORMAT);
		return FALSE;
	}
	Fseek(headlen - 40, fd, SEEK_CUR);

	if (clrused <= 0)
	{
		clrused = 2;
	} else if (clrused > 2)
	{
		Fclose(fd);
		do_alert(AL_UNKNOWN_FORMAT);
		return FALSE;
	}
	
	for (i = 0; i < clrused; i++)
	{
		Fread(fd, 3, palette[i]);
		if (mapentrysize > 3)
			Fseek(1, fd, SEEK_CUR);
		palette[i][3] = 0;
	}
	
	rowsize = width / 8;
	if (topdown)
	{
		data = icondata;
		Fread(fd, rowsize * height, data);
	} else
	{
		data = icondata + rowsize * height;
		for (y = 0; y < height; y++)
		{
			data -= rowsize;
			Fread(fd, rowsize, data);
		}
	}

	Fclose(fd);
	
	if (get_long(palette[0]) != 0xffffffL ||
		get_long(palette[1]) != 0x000000L)
	{
		unsigned long size = rowsize * height;
		data = icondata;
		do
		{
			*data = ~(*data);
			data++;
		} while (--size);
	}

	icon_loaded = TRUE;
		
	return TRUE;
}


static int extract_icon_rom(void)
{
	long romaddr = Supexec(get_rom_addr);
	OSHEADER *hdr = (OSHEADER *)romaddr;
	int id;
	const unsigned char *data;
	
	if (romaddr != 0x0e00000UL)
	{
		if (do_alert(AL_BOGUS_ROMADDR) != 1)
			return FALSE;
	}
	if ((id = check_tosversion(hdr, 0x80000UL)) < 0)
		return FALSE;
	if (!select_file(bmppath, "*.BMP", rs_frstr[FS_BMP]))
		return FALSE;

	data = (const unsigned char *)hdr + versions[id].logoaddr;
	memcpy(icondata, data, ICON_SIZE);
	icon_loaded = TRUE;
	
	write_bmpfile(icondata);
	do_alert(AL_BMP_WRITTEN);
	return TRUE;
}


static unsigned short const crc_tab[256] = {
	0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7, 0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
	0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6, 0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
	0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485, 0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
	0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4, 0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
	0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823, 0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
	0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12, 0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
	0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41, 0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
	0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70, 0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
	0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f, 0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
	0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e, 0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
	0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d, 0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
	0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c, 0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
	0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab, 0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
	0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a, 0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
	0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9, 0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
	0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8, 0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0
};


static unsigned short calc_crc(unsigned long pos, unsigned long size, unsigned long inc)
{
	unsigned short sum;
	unsigned short run;
	unsigned char *p;
	
	p = tos_img + pos;
	sum = run = 0;
	do {
		run = sum;
		sum <<= 8;
		run >>= 8;
		run ^= *p;
		sum ^= crc_tab[run];
		p += inc;
	} while (--size != 0);
	return sum;
}


static int check_rom_crcs(void)
{
	unsigned long size;
	unsigned long banksize;
	unsigned long crcpos;
	unsigned short crc, sum;
	int i, banks;
	
	size = tos_file_size;
	/*
	 * check the current crc.
	 * not all tos versions perform a CRC check,
	 * and the file that we loaded might have been patched already,
	 * so we don't update the crc if it wasn't right before
	 */
	if ((tos_version >= 0x400 && tos_version < 0x500) && size > 256 * ONE_KB)
	{
		/*
		 * Falcon TOS 4.x: a single 512KB ROM
		 */
		banks = 1;
		banksize = ((size + 1) & ~1) - 2;
	} else if ((tos_version == 0x206 || tos_version == 0x207 || tos_version == 0x208) &&
		size > 256 * ONE_KB &&
		!is_emutos)
	{
		/*
		 * ST-Book TOS 2.x: a single 256KB ROM
		 * The second half of the image contains a cartridge ROM
		 */
		banks = 1;
		banksize = (((size >> 1) + 1) & ~1) - 2;
	} else if (size > 256 * ONE_KB)
	{
		/*
		 * TT TOS 3.x: 4 128KB ROMs
		 */
		banks = 4;
		banksize = (((size + 3) / 4) & ~1) - 2;
	} else if (size > 192 * ONE_KB)
	{
		/*
		 * STE TOS 2.x: 2 128KB ROMs
		 */
		banks = 2;
		/*
		 * size should be only -4 (for the 4 checksum bytes at the end),
		 * but i've seen patched TOSes that are 1 byte too short,
		 * hence the +1
		 */
		banksize = (((size + 1) / 2) & ~1) - 2;
	} else
	{
		/*
		 * TOS 1.x: 6 32KB ROMs, without CRC
		 */
		banks = -1;
		banksize = 0;
	}
	if (banks > 0)
	{
		for (i = 0; i < banks; i++)
		{
			crc = calc_crc(i, banksize, banks);
			crcpos = banksize * banks + i;
			sum = (tos_img[crcpos] << 8) | tos_img[crcpos + banks];
			if (crc != sum)
			{
				banks = 0;
				if (do_alert(AL_BADCRC, i) != 1)
					return FALSE;
			}
		}
	}
	return TRUE;
}


static void update_rom_crcs(void)
{
	unsigned long size;
	unsigned long banksize;
	unsigned long crcpos;
	unsigned short crc, sum;
	int i, banks;

	size = tos_file_size;
	/*
	 * check the current crc.
	 * not all tos versions perform a CRC check,
	 * and the file that we loaded might have been patched already,
	 * so we don't update the crc if it wasn't right before
	 */
	if ((tos_version >= 0x400 && tos_version < 0x500) && size > 256 * ONE_KB)
	{
		/*
		 * Falcon TOS 4.x: a single 512KB ROM
		 */
		banks = 1;
		banksize = ((size + 1) & ~1) - 2;
	} else if ((tos_version == 0x206 || tos_version == 0x207 || tos_version == 0x208) &&
		size > 256 * ONE_KB &&
		!is_emutos)
	{
		/*
		 * ST-Book TOS 2.x: a single 256KB ROM
		 * The second half of the image contains a cartridge ROM
		 */
		banks = 1;
		banksize = (((size >> 1) + 1) & ~1) - 2;
	} else if (size > 256 * ONE_KB)
	{
		/*
		 * TT TOS 3.x: 4 128KB ROMs
		 */
		banks = 4;
		banksize = (((size + 3) / 4) & ~1) - 2;
	} else if (size > 192 * ONE_KB)
	{
		/*
		 * STE TOS 2.x: 2 128KB ROMs
		 */
		banks = 2;
		/*
		 * size should be only -4 (for the 4 checksum bytes at the end),
		 * but i've seen patched TOSes that are 1 byte too short,
		 * hence the +1
		 */
		banksize = (((size + 1) / 2) & ~1) - 2;
	} else
	{
		/*
		 * TOS 1.x: 6 32KB ROMs, without CRC
		 */
		banks = -1;
		banksize = 0;
	}
	if (banks > 0)
	{
		int updated = FALSE;

		for (i = 0; i < banks; i++)
		{
			crc = calc_crc(i, banksize, banks);
			crcpos = banksize * banks + i;
			sum = (tos_img[crcpos] << 8) | tos_img[crcpos + banks];
			tos_img[crcpos] = crc >> 8;
			tos_img[crcpos + banks] = crc;
			if (sum != crc)
				updated = TRUE;
		}
		if (updated)
			do_alert(AL_CRC_UPDATED);
	}
}


static int read_tos_img(void)
{
	int fd;
	
	fd = (int)Fopen(tospath, FO_READ);
	if (fd < 0)
	{
		do_alert(AL_FILE_OPEN);
		return FALSE;
	}
	tos_file_size = Fseek(0L, fd, SEEK_END);
	Fseek(0L, fd, SEEK_SET);
	tos_img = Malloc(tos_file_size);
	if (tos_img == NULL)
	{
		do_alert(AL_NOMEM);
		Fclose(fd);
		return FALSE;
	}
	if (Fread(fd, tos_file_size, tos_img) != tos_file_size)
	{
		Mfree(tos_img);
		tos_img = NULL;
		do_alert(AL_READ_ERROR);
		Fclose(fd);
		return FALSE;
	}
	Fclose(fd);
	return TRUE;
}


static int extract_icon_img(void)
{
	int id;
	const unsigned char *data;
	
	if (!select_file(tospath, "*.IMG", rs_frstr[FS_TOS]))
		return FALSE;
	if (!read_tos_img())
		return FALSE;
	if ((id = check_tosversion((OSHEADER *)tos_img, tos_file_size)) < 0)
	{
		Mfree(tos_img);
		tos_img = NULL;
		return FALSE;
	}
	if (!select_file(bmppath, "*.BMP", rs_frstr[FS_BMP]))
	{
		Mfree(tos_img);
		tos_img = NULL;
		return FALSE;
	}
	data = tos_img + versions[id].logoaddr;
	memcpy(icondata, data, ICON_SIZE);
	icon_loaded = TRUE;
	write_bmpfile(icondata);
	Mfree(tos_img);
	tos_img = NULL;
	do_alert(AL_BMP_WRITTEN);
	return TRUE;
}


static int patch_tos(void)
{
	int id;
	int fd;
	
	if (!select_file(tospath, "*.IMG", rs_frstr[FS_TOS]))
		return FALSE;
	if (!read_tos_img())
		return FALSE;
	if ((id = check_tosversion((OSHEADER *)tos_img, tos_file_size)) < 0)
	{
		Mfree(tos_img);
		tos_img = NULL;
		return FALSE;
	}
	if (!select_file(bmppath, "*.BMP", rs_frstr[FS_BMP]))
	{
		Mfree(tos_img);
		tos_img = NULL;
		return FALSE;
	}
	if (!read_bmpfile())
	{
		Mfree(tos_img);
		tos_img = NULL;
		return FALSE;
	}
	memcpy(tos_img + versions[id].logoaddr, icondata, ICON_SIZE);
	update_rom_crcs();
	fd = (int)Fcreate(tospath, 0);
	if (fd < 0)
	{
		do_alert(AL_FILE_CREATE);
		Mfree(tos_img);
		tos_img = NULL;
		return FALSE;
	}
	if (Fwrite(fd, tos_file_size, tos_img) != tos_file_size)
	{
		Mfree(tos_img);
		tos_img = NULL;
		do_alert(AL_WRITE_ERROR);
		Fclose(fd);
		return FALSE;
	}
	Mfree(tos_img);
	tos_img = NULL;
	Fclose(fd);
	do_alert(AL_TOS_PATCHED);
	return TRUE;
}


static int convert_bmp2abs(void)
{
	char *p;
	int fd;
	
	if (!select_file(bmppath, "*.BMP", rs_frstr[FS_BMP]))
		return FALSE;
	if (!read_bmpfile())
		return FALSE;
	strcpy(abspath, bmppath);
	p = strrchr(abspath, '.');
	if (p == NULL)
		strcat(abspath, ".ABS");
	else
		strcpy(p, ".ABS");
	fd = (int)Fcreate(abspath, 0);
	if (fd < 0)
	{
		do_alert(AL_FILE_CREATE);
		return FALSE;
	}
	Fwrite(fd, ICON_SIZE, icondata);
	Fclose(fd);
	do_alert(AL_ABS_WRITTEN);
	return TRUE;
}


static int convert_abs2bmp(void)
{
	char *p;
	int fd;
	long filesize;
	
	if (!select_file(abspath, "*.ABS", rs_frstr[FS_ABS]))
		return FALSE;
	fd = (int)Fopen(abspath, 0);
	if (fd < 0)
	{
		do_alert(AL_FILE_OPEN);
		return FALSE;
	}
	filesize = Fseek(0L, fd, SEEK_END);
	if (filesize != ICON_SIZE)
	{
		Fclose(fd);
		do_alert(AL_UNSUPPORTED_FORMAT);
		return FALSE;
	}
	Fseek(0L, fd, SEEK_SET);
	Fread(fd, ICON_SIZE, icondata);
	Fclose(fd);
	strcpy(bmppath, abspath);
	p = strrchr(bmppath, '.');
	if (p == NULL)
		strcat(abspath, ".BMP");
	else
		strcpy(p, ".BMP");
	if (!write_bmpfile(icondata))
		return FALSE;
	do_alert(AL_BMP_WRITTEN);

	icon_loaded = TRUE;
	return TRUE;
}


static void main_dialog(void)
{
	_WORD obj;
	
	do
	{
#if 0
		OBJECT *tree = rs_trindex[MAIN];
		
		if (icon_loaded)
			tree[PATCH_TOS].ob_state &= ~OS_DISABLED;
		else
			tree[PATCH_TOS].ob_state |= OS_DISABLED;
#endif
		obj = do_dialog(MAIN, 0, 0);
		switch (obj)
		{
		case MAIN_CANCEL:
			obj = -1;
			break;
		case EXTRACT_ICON_IMG:
			extract_icon_img();
			break;
		case PATCH_TOS:
			patch_tos();
			break;
		case EXTRACT_ICON_ROM:
			extract_icon_rom();
			break;
		case CONVERT_BMP2ABS:
			convert_bmp2abs();
			break;
		case CONVERT_ABS2BMP:
			convert_abs2bmp();
			break;
		}
	} while (obj > 0);
}


int main(void)
{
	BITBLK *bi;
	
	if (appl_init() < 0)
		return 1;
	graf_mouse(ARROW, NULL);
	
	aes_handle = graf_handle(&gl_wchar, &gl_hchar, &dummy, &dummy);
	tos_fuji_rsc_load(gl_wchar, gl_hchar);
	bi = rs_trindex[MAIN][FUJI].ob_spec.bitblk;
	bi->bi_wb = ICON_WIDTH / 8;
	bi->bi_hl = ICON_HEIGHT;
	bi->bi_pdata = (void *)icondata;
	main_dialog();
		
	appl_exit();
	return 0;
}
